"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""


class BotChannelMapper(object):
    """
    Class that maps bots with non-standard whois responses to a channel
    """

    bot_channel_map = {"HelloKitty": "#horriblesubs"}

    @staticmethod
    def has_mapping(botname: str) -> bool:
        """
        Checks if a bot is mapped to a channel in this class

        :param botname: The bot's name
        :return:        True, if it has a mapping, false otherwise
        """
        return botname in BotChannelMapper.bot_channel_map.keys()

    @staticmethod
    def map(botname: str) -> str:
        """
        Maps a bot to a channel name

        :param botname: The bot's name
        :return:        The mapped channel name
        """
        return BotChannelMapper.bot_channel_map[botname]



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import time
import urwid
from threading import Thread
from xdcc_dl.metadata import General
from xdcc_dl.entities.Progress import Progress
from xdcc_dl.pack_searchers.PackSearcher import PackSearcher
from xdcc_dl.entities.XDCCPack import xdcc_packs_from_xdcc_message
from xdcc_dl.xdcc.MultipleServerDownloader import MultipleServerDownloader


# noinspection PyUnusedLocal
class XDCCDownloaderTui(object):
    """
    Class that models a urwid TUI for the XDCC Downloader
    """

    def __init__(self) -> None:
        """
        Initializes the TUI's various widgets
        """
        self.upper_body = []
        self.upper_middle_body = []
        self.middle_body = []
        self.lower_middle_body = []
        self.lower_body = []

        self.loop = None
        self.top = None
        self.list_walker = None

        self.downloading = False
        self.searching = False

        self.downloader = None

        self.search_results = []
        self.search_results_checks = []
        self.download_queue = []
        self.download_queue_checks = []

        self.gpl_notice = urwid.Text("XDCC Downloader V " + General.version_number + "\n"
                                     "Copyright (C) 2016 Hermann Krumrey\n\n"
                                     "This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'.\n"
                                     "This is free software, and you are welcome to redistribute it\n"
                                     "under certain conditions; type `show c' for details.")
        self.message_edit = urwid.Edit(caption="XDCC Message: ", edit_text="")
        self.server_edit = urwid.Edit(caption="Server: ", edit_text="irc.rizon.net")
        self.add_pack_button = urwid.Button("Add Pack")

        self.search_engine_label = urwid.Text("Search Engines")
        self.search_engine_options = []
        for search_engine in ["All"] + PackSearcher.get_available_pack_searchers():
            urwid.RadioButton(self.search_engine_options, search_engine)

        self.search_term_edit = urwid.Edit(caption="Search Term: ", edit_text="")
        self.search_button = urwid.Button("Search")
        self.search_results_label = urwid.Text("Search Results:")
        self.add_search_result_button = urwid.Button("Add Selected Packs")

        self.download_queue_label = urwid.Text("Download_Queue")
        self.remove_pack_button = urwid.Button("Remove Selected Packs")

        self.destination_edit = urwid.Edit(caption="Destination Directory: ", edit_text=os.getcwd())
        self.download_button = urwid.Button("Download")
        self.single_progress_bar = urwid.ProgressBar("Progress", "Total")
        self.total_progress_bar = urwid.ProgressBar("Progress", "Total")

        self.connect_widgets()
        self.lay_out()

    def connect_widgets(self) -> None:
        """
        Connects actions to the relevant widgets

        :return: None
        """
        urwid.connect_signal(self.add_pack_button, 'click', self.add_pack_manually)
        urwid.connect_signal(self.search_button, 'click', self.search)
        urwid.connect_signal(self.add_search_result_button, 'click', self.add_selected_search_result_packs)
        urwid.connect_signal(self.remove_pack_button, 'click', self.remove_selected_packs)
        urwid.connect_signal(self.download_button, 'click', self.download)

    def lay_out(self) -> None:
        """
        Handles the layout of the TUI elements

        :return: None
        """
        div = urwid.Divider()

        self.upper_body = [self.gpl_notice, div, self.message_edit, self.server_edit, self.add_pack_button, div,
                           self.search_engine_label] + self.search_engine_options +\
                          [div, self.search_term_edit, self.search_button, div, self.search_results_label]
        self.upper_middle_body = []
        self.middle_body = [self.add_search_result_button, div, self.download_queue_label]
        self.lower_middle_body = []
        self.lower_body = [self.remove_pack_button, div, self.destination_edit, self.download_button, div,
                           self.single_progress_bar, self.total_progress_bar]

        body = self.upper_body + self.upper_middle_body + self.middle_body + self.lower_middle_body + self.lower_body

        self.list_walker = urwid.SimpleFocusListWalker(body)
        self.top = urwid.Overlay(urwid.Padding(urwid.ListBox(self.list_walker), left=2, right=2),
                                 urwid.SolidFill(u'\N{MEDIUM SHADE}'),
                                 align='center', width=('relative', 80),
                                 valign='middle', height=('relative', 70),
                                 min_width=20, min_height=10)

    def start(self) -> None:  # pragma: no cover
        """
        Starts the TUI

        :return: None
        """
        self.loop = urwid.MainLoop(self.top, palette=[('reversed', 'standout', '')])

        try:
            self.loop.run()
        except Exception as e:
            if self.downloader is not None:
                self.downloader.quit()
            raise e

    def refresh_ui(self) -> None:
        """
        Refreshed the UI

        :return: None
        """
        self.download_queue_checks = []
        self.search_results_checks = []

        for result in self.search_results:
            self.search_results_checks.append(urwid.CheckBox(result.get_filename()))
        for item in self.download_queue:
            self.download_queue_checks.append(urwid.CheckBox(item.get_request_message(full=True)))

        self.upper_middle_body = self.search_results_checks
        self.lower_middle_body = self.download_queue_checks

        body = self.upper_body + self.upper_middle_body + self.middle_body + self.lower_middle_body + self.lower_body

        self.list_walker[:] = body
        self.loop.draw_screen()

    def add_pack_manually(self, button: urwid.Button) -> None:
        """
        Adds a manually defined XDCC pack to the download queue

        :param button: The Button that called this method
        :return:       None
        """
        message = self.message_edit.get_edit_text()
        server = self.server_edit.get_edit_text()
        self.download_queue += xdcc_packs_from_xdcc_message(message, server=server)
        self.refresh_ui()

    def add_selected_search_result_packs(self, button: urwid.Button) -> None:
        """
        Adds all selected packs in the search results to the download queue

        :param button: The Button that called this method
        :return:       None
        """
        for index, result in enumerate(self.search_results_checks):
            if result.get_state():
                self.download_queue.append(self.search_results[index])
        self.refresh_ui()

    def remove_selected_packs(self, button: urwid.Button) -> None:
        """
        Removes all selected packs from the download queue

        :param button: The Button that called this method
        :return:       None
        """
        pop_indexes = []
        for index, pack in enumerate(self.download_queue_checks):
            if pack.get_state():
                pop_indexes.append(index)

        for index in reversed(sorted(pop_indexes)):
            self.download_queue.pop(index)

        self.refresh_ui()

    def search(self, button: urwid.Button) -> None:
        """
        Starts searching with the configured options

        :param button: The Button that called this method
        :return:       None
        """
        if not self.searching:

            self.searching = True

            search_engine = list(filter(lambda x: x.get_state(), self.search_engine_options))[0].get_label()
            searcher = PackSearcher() if search_engine == "All" else PackSearcher([search_engine])
            search_term = self.search_term_edit.get_edit_text()

            def do_search():
                self.spin_buttons(search=True)
                self.search_results = searcher.search(search_term)
                self.refresh_ui()
                self.searching = False

            Thread(target=do_search).start()

    def download(self, button: urwid.Button) -> None:
        """
        Starts downloading all packs that are currently in the download queue

        :param button: The Button that called this method
        :return:       None
        """
        if not self.downloading:
            self.downloading = True

            destination = self.destination_edit.get_edit_text()
            if not os.path.isdir(destination):
                self.downloading = False
                self.show_message_popup("The entered directory does not exist.")
                return

            for pack in self.download_queue:
                pack.set_directory(destination)

            def do_download():

                progress = Progress(len(self.download_queue),
                                    callback=lambda a, b, single_percentage, d, e, total_percentage, g, h:
                                    self.update_progress(single_percentage, total_percentage))

                self.spin_buttons(download=True)
                self.downloader = MultipleServerDownloader("random")
                self.downloader.download(self.download_queue, progress)
                self.download_queue = []
                self.update_progress(0.0, 0.0)
                self.refresh_ui()
                self.downloader.quit()
                self.downloading = False

            Thread(target=do_download).start()

    def update_progress(self, single_precentage: float, total_percentage: float) -> None:
        """
        Updates the progress bars

        :param single_precentage: The single completion so far
        :param total_percentage:  The total completion so far
        :return:                  None
        """
        self.single_progress_bar.set_completion(single_precentage)
        self.total_progress_bar.set_completion(total_percentage)
        self.loop.draw_screen()

    def spin_buttons(self, download: bool = False, search: bool = False) -> None:
        """
        Starts the button spin animations

        :param download: If set, starts the spin animation for the download button
        :param search:   If set, starts the spin animation for the search button
        :return:         None
        """

        def spin_thread():

            while self.downloading or self.searching:

                if self.downloading and download:
                    new_text = "Downloading" + (self.download_button.get_label().count(".") % 3 + 1) * "."
                    self.download_button.set_label(new_text)

                if self.searching and search:
                    new_text = "Searching" + (self.search_button.get_label().count(".") % 3 + 1) * "."
                    self.search_button.set_label(new_text)

                self.loop.draw_screen()
                time.sleep(0.3)

            if download:
                self.download_button.set_label("Download")
            if search:
                self.search_button.set_label("Search")
            self.loop.draw_screen()

        Thread(target=spin_thread).start()

    def show_message_popup(self, message: str) -> None:
        """
        Method that shows a message popup dialog while hiding the TUI

        :param message: The message to display
        :return:        None
        """

        text = urwid.Text(message)
        button = urwid.Button("OK")
        urwid.connect_signal(button, 'click', lambda x: self.refresh_ui())  # pragma: no cover

        self.list_walker[:] = [text, button]
        self.loop.draw_screen()



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import re
from typing import List
from xdcc_dl.entities.IrcServer import IrcServer


class XDCCPack(object):
    """
    Class that models an XDCC Pack
    """

    def __init__(self, server: IrcServer, bot: str, packnumber: int) -> None:
        """
        Initializes an XDCC object. It contains all the necessary information for joining the correct
        IRC server and channel and sending the download request to the correct bot, then storing the
        received file in the predetermined location. If the destination is a directory, the file will be stored
        in the directory with the default file name, if not the file will be saved at the destination exactly.
        The file extension will stay as in the original filename

        :param server:       The Sever to be used by the XDCC Bot
        :param bot:          The bot serving the file
        :param packnumber:   The packnumber of the desired file
        """
        self.server = server
        self.bot = bot
        self.packnumber = packnumber
        self.directory = os.getcwd()
        self.filename = ""
        self.size = 0

        self.original_filename = ""

    def is_filename_valid(self, filename: str) -> bool:
        """
        Checks if a filename is the same as the original filename, if one was set previously.
        This is used internally by the IRC Bot to check if a file that was offered to the bot actually matches
        the file we want to download.

        :param filename: The file name to check
        :return:         True, if the names match, or no original filename was set, otherwise False
        """
        return filename == self.original_filename if self.original_filename != "" else True

    def set_filename(self, filename: str, override: bool = False) -> None:
        """
        Sets the filename (or only the file extension) of the target file

        :param filename: the filename as provided by the XDCC bot
        :param override: Overrides the current filename
        :return:         None
        """
        if self.filename and len(filename.split(".")) > 1 and not override:
            extension = filename.rsplit(".", 1)[1]
            if not self.filename.endswith(extension):
                self.filename += "." + extension

        if not self.filename or override:
            self.filename = filename

    def set_original_filename(self, filename: str) -> None:
        """
        Sets the 'original' filename, a.k.a the name of the actual file to download.
        This is a method that should only be used by the pack searchers to add filename checks
        during the download.

        :param filename: The original filename as found by the PackSearcher
        :return:         None
        """
        self.original_filename = filename

    def set_directory(self, directory: str) -> None:
        """
        Sets the target directory of the XDCC PAck

        :param directory: the target directory
        :return:          None
        """
        self.directory = directory

    def set_size(self, size: int) -> None:
        """
        Sets the file size of the XDCC pack

        :param size: the size of the pack
        :return:     None
        """
        self.size = size

    def get_server(self) -> IrcServer:
        """
        :return: The server
        """
        return self.server

    def get_filepath(self) -> str:
        """
        :return: The full destination file path
        """
        return os.path.join(self.directory, self.filename)

    def get_filename(self) -> str:
        """
        :return: The currently set filename
        """
        return self.filename

    def get_size(self) -> int:
        """
        :return: The currently set file size
        """
        return self.size

    def get_bot(self) -> str:
        """
        :return: The bot
        """
        return self.bot

    def get_packnumber(self) -> int:
        """
        :return: the pack number
        """
        return self.packnumber

    def get_request_message(self, full: bool = False) -> str:
        """
        Generates an xdcc send message to be sent to the bot to initiate the XDCC connection

        :param full: Returns the entire message string, including the bot's name, as seen on packlist sites
        :return:     The generated message string
        """
        if full:
            return "/msg " + self.bot + " xdcc send #" + str(self.packnumber)
        else:
            return "xdcc send #" + str(self.packnumber)


def xdcc_packs_from_xdcc_message(xdcc_message: str,
                                 destination_directory: str = os.getcwd(),
                                 server: str = "irc.rizon.net") -> List[XDCCPack]:
    """
    Generates XDCC Packs from an xdcc message of the form "/msg <bot> xdcc send #<packnumber>[-<packnumber>]"

    :param xdcc_message:           the XDCC message to parse
    :param destination_directory:  the destination directory of the file
    :param server:                 the server to use, defaults to irc.rizon.net for simplicity's sake
    :return:                       The generated XDCC Packs in a list
    """
    if not re.search(r"^/msg [^ ]+ xdcc send #[0-9]+(-[0-9]+(,[0-9]+)?)?$", xdcc_message):
        return []

    bot = xdcc_message.split("/msg ")[1].split(" ")[0]

    try:
        packnumber = int(xdcc_message.rsplit("#", 1)[1])
        xdcc_pack = XDCCPack(IrcServer(server), bot, packnumber)
        xdcc_pack.set_directory(destination_directory)
        return [xdcc_pack]

    except ValueError:
        packnumbers = xdcc_message.rsplit("#", 1)[1]
        start, end = packnumbers.split("-")

        try:
            step = int(end.split(",")[1])
            end = end.split(",")[0]
        except (IndexError, ValueError):
            step = 1

        packs = []
        for pack in range(int(start), int(end) + 1, step):
            xdcc_pack = XDCCPack(IrcServer(server), bot, pack)
            xdcc_pack.set_directory(destination_directory)
            packs.append(xdcc_pack)
        return packs



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import unittest
from xdcc_dl.entities.IrcServer import IrcServer
from xdcc_dl.entities.XDCCPack import XDCCPack, xdcc_packs_from_xdcc_message


class UnitTests(unittest.TestCase):

    def setUp(self):
        self.pack = XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 1)

    def tearDown(self):
        pass

    def test_getters(self):

        self.assertEqual(self.pack.get_packnumber(), 1)
        self.assertEqual(self.pack.get_bot(), "xdcc_servbot")
        self.assertEqual(self.pack.get_filename(), "")
        self.assertEqual(self.pack.get_size(), 0)
        self.assertEqual(self.pack.get_server().get_address(), "irc.namibsun.net")
        self.assertEqual(self.pack.get_filepath(), os.getcwd() + os.sep)
        self.assertEqual(self.pack.get_request_message(), "xdcc send #1")

    def test_setting_filename(self):

        self.pack.set_filename("test")
        self.assertEqual(self.pack.get_filename(), "test")
        self.assertEqual(self.pack.get_filepath(), os.path.join(os.getcwd(), "test"))

        self.pack.set_filename("something")
        self.assertEqual(self.pack.get_filename(), "test")
        self.assertEqual(self.pack.get_filepath(), os.path.join(os.getcwd(), "test"))

        self.pack.set_filename("something", override=True)
        self.assertEqual(self.pack.get_filename(), "something")
        self.assertEqual(self.pack.get_filepath(), os.path.join(os.getcwd(), "something"))

        self.pack.set_filename("something_else.txt")
        self.assertEqual(self.pack.get_filename(), "something.txt")
        self.assertEqual(self.pack.get_filepath(), os.path.join(os.getcwd(), "something.txt"))

        self.pack.set_filename("something_else.txt")
        self.assertEqual(self.pack.get_filename(), "something.txt")
        self.assertEqual(self.pack.get_filepath(), os.path.join(os.getcwd(), "something.txt"))

        self.pack.set_filename("something_else.mkv")
        self.assertEqual(self.pack.get_filename(), "something.txt.mkv")
        self.assertEqual(self.pack.get_filepath(), os.path.join(os.getcwd(), "something.txt.mkv"))

        self.pack.set_directory(os.path.join(os.getcwd(), "test"))
        self.assertEqual(self.pack.get_filename(), "something.txt.mkv")
        self.assertEqual(self.pack.get_filepath(), os.path.join(os.getcwd(), "test", "something.txt.mkv"))

    def test_original_filename_check(self):

        self.assertTrue(self.pack.is_filename_valid("the_original"))
        self.assertTrue(self.pack.is_filename_valid("not_the_original"))
        self.pack.set_original_filename("the_original")
        self.assertTrue(self.pack.is_filename_valid("the_original"))
        self.assertFalse(self.pack.is_filename_valid("not_the_original"))

    def test_request_message(self):
        self.assertEqual(self.pack.get_request_message(), "xdcc send #1")
        self.assertEqual(self.pack.get_request_message(full=True), "/msg xdcc_servbot xdcc send #1")

    def test_generating_from_xdcc_message_single(self):

        packs = xdcc_packs_from_xdcc_message("/msg xdcc_servbot xdcc send #1", "testdir", "irc.namibsun.net")
        self.assertEqual(len(packs), 1)
        pack = packs[0]

        self.assertEqual(pack.get_packnumber(), 1)
        self.assertEqual(pack.get_bot(), "xdcc_servbot")
        self.assertEqual(pack.get_server().get_address(), "irc.namibsun.net")
        self.assertEqual(pack.get_filepath(), "testdir" + os.sep)
        self.assertTrue(pack.get_request_message() in "/msg xdcc_servbot xdcc send #1")

    def test_generating_from_xdcc_message_range(self):

        packs = xdcc_packs_from_xdcc_message("/msg xdcc_servbot xdcc send #1-100")
        self.assertEqual(len(packs), 100)

        for i, pack in enumerate(packs):
            self.assertEqual(pack.get_packnumber(), i + 1)
            self.assertEqual(pack.get_server().get_address(), "irc.rizon.net")
            self.assertEqual(pack.get_filepath(), os.getcwd() + os.sep)

    def test_generating_from_xdcc_message_range_with_jumps(self):

        packs = xdcc_packs_from_xdcc_message("/msg xdcc_servbot xdcc send #1-100,2")
        self.assertEqual(len(packs), 50)

        i = 1
        for pack in packs:
            self.assertEqual(pack.get_packnumber(), i)
            i += 2

    def test_generating_invalid_xdcc_message(self):

        packs = xdcc_packs_from_xdcc_message("randomnonesense")
        self.assertEqual(len(packs), 0)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

"""
The metadata is stored here. It can be used by any other module in this project this way, most
notably by the setup.py file
"""


class GitRepository:  # pragma: no cover
    """
    Class that stores information about the git repository sites used by this project
    """

    repository_name = "xdcc-downloader"
    """
    The name of the repository
    """

    github_owner = "namboy94"
    """
    The owner's Github username
    """

    gitlab_owner = "namboy94"
    """
    The project's owner's username on Gitlab
    """

    gitlab_site_url = "https://gitlab.namibsun.net/"
    """
    The address of the Gitlab instance
    """

    github_url = "https://github.com/" + github_owner + "/" + repository_name
    """
    The Github site URL
    """

    gitlab_url = gitlab_site_url + gitlab_owner + "/" + repository_name
    """
    The Gitlab Project URL
    """


class General:  # pragma: no cover
    """
    Class that stores general information about a project
    """

    project_description = "An XDCC Downloader"
    """
    A short description of the project
    """

    version_number = "1.1.2"
    """
    The current version of the program.
    """

    author_names = "Hermann Krumrey"
    """
    The name(s) of the project author(s)
    """

    author_emails = "hermann@krumreyh.com"
    """
    The email address(es) of the project author(s)
    """

    license_type = "GNU GPL3"
    """
    The project's license type
    """

    project_name = GitRepository.repository_name
    """
    The name of the project
    """

    download_master_zip = GitRepository.gitlab_url + "/repository/archive.zip?ref=master"
    """
    A URL linking to the current source zip file of the master branch.
    """


class PypiVariables:  # pragma: no cover
    """
    Variables used for distributing with setuptools to the python package index
    """

    classifiers = [

        "Environment :: Console",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Development Status :: 1 - Planning",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
        "Topic :: Communications :: File Sharing",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"

    ]
    """
    The list trove classifiers applicable to this project
    """

    install_requires = ["raven", "irc", "bs4", "requests", "cfscrape", "urwid", "typing"]
    """
    Python Packaging Index dependencies
    """

    name = "xdcc_dl"
    """
    The name of the project on Pypi
    """

    version = General.version_number
    """
    The version of the project on pypi
    """

    description = General.project_description
    """
    The short description of the project on pypi
    """

    url = GitRepository.gitlab_url
    """
    A URL linking to the home page of the project, in this case a
    self-hosted Gitlab page
    """

    download_url = General.download_master_zip
    """
    A link to the current source zip of the project
    """

    author = General.author_names
    """
    The author(s) of this project
    """

    author_email = General.author_emails
    """
    The email adress(es) of the author(s)
    """

    license = General.license_type
    """
    The License used in this project
    """


class SentryLogger:  # pragma: no cover
    """
    Class that handles the sentry logger initialization
    """

    sentry_dsn = "https://3f4217fbc10a48bf8bb119c1782d8b03:58b2a299d71d4c36a277df9add7b38c3@sentry.io/110685"
    """
    The DSN associated with this project
    """

    sentry = None
    """
    The sentry client
    """

    # Create the Sentry client to log bugs
    try:
        from raven import Client
        sentry = Client(dsn=sentry_dsn, release=General.version_number)
    except ImportError:
        Client = None



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import irc.client
# noinspection PyPep8Naming
from xdcc_dl.logging.LoggingTypes import LoggingTypes as LOG
from xdcc_dl.xdcc.layers.irc.IrcEventPrinter import IrcEventPrinter
from xdcc_dl.xdcc.layers.helpers.BotChannelMapper import BotChannelMapper


class BotNotFoundException(Exception):
    """
    An Exception to throw for when the WHOIS query does not return any results
    """
    pass


# noinspection PyUnusedLocal
class BotFinder(IrcEventPrinter):
    """
    Class that uses WHOIS queries to find and joins the channels a bot is a part of
    Layer 2 of the XDCC Bot
    """

    def on_welcome(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        The Welcome Event indicates that the server connection was established.
        A whois is sent to figure out which channels the bot resides in

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        # noinspection PyUnresolvedReferences
        super().on_welcome(connection, event)
        self.logger.log("Sending WHOIS command for " + self.current_pack.get_bot(), LOG.WHOIS_SEND)
        connection.whois(self.current_pack.get_bot())
        # -> Success: on_whoischannels or on_endofwhois
        # -> Failure: on_nosuchnick
        # Additional Effects:  ~> on_whoisuser
        #                      ~> on_whoisserver

    def on_whoischannels(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        A successful WHOIS query will result in this method being called. The bot will then attempt to join all
        channels the bot also joined

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.channel_join_required = True  # Some bots are not on any channel, and can be used
        # without joining any channel. If a whoischannels message is
        # received however, we need to join a channel

        self.logger.log("Received WHOIS information, bot resides in: " + event.arguments[1], LOG.WHOIS_SUCCESS)
        channels = event.arguments[1].split("#")
        channels.pop(0)

        new_channel_joined = False

        for channel in channels:
            channel = "#" + channel.split(" ")[0]

            if channel not in self.joined_channels:
                new_channel_joined = True
                self.joined_channels.append(channel)
                self.logger.log("Joining Channel " + channel, LOG.CHANNEL_JOIN_ATTEMPT)
                connection.join(channel)  # -> on_join

        if not new_channel_joined:
            self.logger.log("Channels joined already", LOG.CHANNEL_JOIN_SUCCESS)
            event.source = self.user.get_name()
            # noinspection PyUnresolvedReferences
            self.on_join(connection, event)

    def on_endofwhois(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Checks the end of a WHOIS command if a channel join has occured or was even necessary
        If it was not necessary, starts the download

        :param connection: the IRC connection
        :param event:      the endofwhois event
        :return:           None
        """
        if not self.channel_join_required:

            if BotChannelMapper.has_mapping(self.current_pack.get_bot()):
                event.arguments = [None, BotChannelMapper.map(self.current_pack.get_bot())]
                self.on_whoischannels(connection, event)

            else:
                event.source = self.user.get_name()
                # noinspection PyUnresolvedReferences
                self.on_join(connection, event)  # Simulates a Channel Join if joining a channel is unnecessary
                # -> on_join

    def on_nosuchnick(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        This method is called if the WHOIS query fails, i.e. the bot does not exist on the IRC server
        It will forcefully abort the connection, which will then result in the bot skipping the current Pack

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        if event.arguments[0] == self.current_pack.get_bot():  # Make sure the failed WHOIS is for our bot
            self.logger.log("Bot " + self.current_pack.get_bot() + " does not exist on Server", LOG.WHOIS_NO_RESULT)
            raise BotNotFoundException()

    def on_whoisuser(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Method called when a whoisuser command was received

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        message = ""
        for part in event.arguments:
            message += part + " "
        message = message.lstrip().rstrip()
        self.logger.log(message, LOG.WHOIS_USER)

    def on_whoisserver(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Method called when a whoisserver command was received

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        message = ""
        for part in event.arguments:
            message += part + " "
        message = message.lstrip().rstrip()
        self.logger.log(message, LOG.WHOIS_SERVER)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import time
import unittest
from xdcc_dl.tui.XDCCDownloaderTui import XDCCDownloaderTui
from xdcc_dl.entities.XDCCPack import xdcc_packs_from_xdcc_message


class LoopDummy(object):
    def draw_screen(self):
        pass


# noinspection PyTypeChecker
class UnitTests(unittest.TestCase):

    def setUp(self):
        self.tui = XDCCDownloaderTui()
        self.tui.loop = LoopDummy()

    def tearDown(self):
        if os.path.isfile("1_test.txt"):
            os.remove("1_test.txt")
        if os.path.isfile("2_test.txt"):
            os.remove("2_test.txt")
        if os.path.isfile("3_test.txt"):
            os.remove("3_test.txt")

    def test_adding_pack(self):
        self.assertEqual(len(self.tui.download_queue), 0)
        self.tui.message_edit.set_edit_text("/msg xdcc_servbot xdcc send #1")
        self.tui.server_edit.set_edit_text("irc.namibsun.net")
        self.tui.add_pack_manually(None)
        self.assertEqual(len(self.tui.download_queue), 1)

    def test_searching(self):
        self.assertEqual(len(self.tui.search_results), 0)
        self.tui.search_term_edit.set_edit_text("1_test.txt")
        self.select_search_engine("namibsun")
        self.tui.search(None)

        while self.tui.searching:
            pass

        self.assertEqual(len(self.tui.search_results), 1)

    def test_searching_while_searching(self):
        self.assertEqual(len(self.tui.search_results), 0)
        self.tui.search_term_edit.set_edit_text("one punch man")
        self.select_search_engine("nibl")
        self.tui.search(None)

        self.tui.search_term_edit.set_edit_text("1_test.txt")
        self.select_search_engine("namibsun")
        self.tui.search(None)

        while self.tui.searching:
            pass

        self.assertLess(10, len(self.tui.search_results))
        for pack in self.tui.search_results:
            self.assertNotEqual("1_test.txt", pack.get_filename())

    def test_adding_search_results_to_queue(self):
        self.test_searching()

        self.assertEqual(0, len(self.tui.download_queue))
        self.tui.add_selected_search_result_packs(None)
        self.tui.search_results_checks[0].set_state(True)
        self.assertEqual(0, len(self.tui.download_queue))
        self.tui.add_selected_search_result_packs(None)
        self.assertEqual(1, len(self.tui.download_queue))

    def test_removing_packs_from_queue(self):
        self.tui.download_queue = xdcc_packs_from_xdcc_message("/msg xdcc_servbot xdcc send #1-3")
        self.assertEqual(3, len(self.tui.download_queue))
        self.tui.refresh_ui()
        self.tui.download_queue_checks[0].set_state(True)
        self.tui.remove_selected_packs(None)
        self.assertEqual(2, len(self.tui.download_queue))

    def test_download(self):
        self.tui.download_queue = xdcc_packs_from_xdcc_message("/msg xdcc_servbot xdcc send #1-3",
                                                               server="irc.namibsun.net")
        self.tui.refresh_ui()
        self.tui.download(None)

        while self.tui.downloading:
            pass

        self.assertTrue(os.path.isfile("1_test.txt"))
        self.assertTrue(os.path.isfile("2_test.txt"))
        self.assertTrue(os.path.isfile("3_test.txt"))

    def test_download_while_downloading(self):
        self.tui.download_queue = xdcc_packs_from_xdcc_message("/msg xdcc_servbot xdcc send #1-3",
                                                               server="irc.namibsun.net")
        self.tui.refresh_ui()
        self.tui.download(None)

        time.sleep(0.5)

        self.assertNotEqual(self.tui.download_button.get_label(), "Download")

        self.tui.download_queue = xdcc_packs_from_xdcc_message("/msg Ginpachi-Sensei xdcc send #1")
        self.tui.refresh_ui()
        self.tui.download(None)

        while self.tui.downloading:
            pass

        self.assertTrue(os.path.isfile("1_test.txt"))
        self.assertTrue(os.path.isfile("2_test.txt"))
        self.assertTrue(os.path.isfile("3_test.txt"))
        self.assertFalse(os.path.isfile("Gin.txt"))

    def test_download_wrong_directory(self):
        self.tui.download_queue = xdcc_packs_from_xdcc_message("/msg xdcc_servbot xdcc send #1-3",
                                                               server="irc.namibsun.net")

        self.assertEqual(self.tui.destination_edit.get_edit_text(), os.getcwd())
        self.tui.destination_edit.set_edit_text("NotADirectory")
        self.assertEqual(self.tui.destination_edit.get_edit_text(), "NotADirectory")

        self.tui.refresh_ui()
        self.tui.download(None)

        while self.tui.downloading:
            pass

        self.assertFalse(os.path.isfile("1_test.txt"))
        self.assertFalse(os.path.isfile("2_test.txt"))
        self.assertFalse(os.path.isfile("3_test.txt"))

    def select_search_engine(self, option):
        for engine in self.tui.search_engine_options:
            if engine.get_label() == option:
                engine.set_state(True)
            else:
                engine.set_state(False)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import irc.client
import irc.events
# noinspection PyUnresolvedReferences,PyPep8Naming
from xdcc_dl.logging.LoggingTypes import LoggingTypes as LOG
from xdcc_dl.xdcc.layers.irc.BaseIrcClient import BaseIrclient


def create_event_printer(event: str, formatting: str, show_tag: bool = False) -> str:
    """
    Creates a new method definition for an on_event method that can then be integrated using exec

    :param event:      The name of the event on which this method is called
    :param formatting: The formatting from LOG, but as a string
    :param show_tag:   Determines if an 'on_event" string should be printed beforehand
    :return:           The method as an executable string
    """
    method_definition = "def on_" + event + "(self, connection, event):\n"
    if show_tag:
        method_definition += "    self.logger.log(\"on_" + event + "\", LOG.EVENT)\n"
    method_definition += "    print_args = \"\"\n"
    method_definition += "    for arg in event.arguments:\n"
    method_definition += "        print_args += arg + \" \"\n"
    method_definition += "    print_args = print_args.rstrip().lstrip()\n"
    method_definition += "    self.logger.log(print_args, " + formatting + ")"
    return method_definition


# noinspection PyMethodMayBeStatic,PyUnusedLocal
class IrcEventPrinter(BaseIrclient):
    """
    Class that prints output to the console for every type of IRC event defined by the irc library
    Layer 1 of the XDCC Bot
    """

    # This assigns a method to all event types that are supported by the IRC library.
    # They print the type of event and its parameters
    for event in irc.events.all:
        if event != "disconnect":
            method = create_event_printer(event, "LOG.EVENT_TEXT", True)
            exec(method)

    # Here are some more specialized ones

    # Prints the server greeting
    for event in ["welcome", "yourhost", "created", "myinfo", "featurelist", "luserclient", "luserop",
                  "luserunknown", "luserchannels", "luserme", "n_local", "n_global", "modstart"]:
        exec(create_event_printer(event, "LOG.WELCOME", False))

    # Message of the Day
    for event in ["motd", "motdstart", "endofmotd"]:
        exec(create_event_printer(event, "LOG.MESSAGE_OF_THE_DAY", False))

    # Here we manually define the more interesting methods:

    def on_privnotice(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Prints Private Notices

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log(event.arguments[0], LOG.PRIVATE_NOTICE)

    def on_privmsg(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Prints Private Messages

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log(event.source + ": " + event.arguments[0], LOG.PRIVATE_MESSAGE)

    def on_pubnotice(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Prints Public Notices

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log(event.source + ": " + event.arguments[0], LOG.PUBLIC_NOTICE)

    def on_pubmsg(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Prints Public Messages

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log(event.source + ": " + event.arguments[0], LOG.PUBLIC_MESSAGE)

    def on_ping(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Prints Pings

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log("PING: " + event.arguments[0], LOG.PING)

    def on_ctcp(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Prints Client-To-Client_Protocol 'Version' Messages

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        if event.arguments[0] == "VERSION":
            self.logger.log(event.arguments[0], LOG.CTCP_VERSION)
            return



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import unittest
from xdcc_dl.xdcc.XDCCDownloader import XDCCDownloader
from xdcc_dl.pack_searchers.PackSearcher import PackSearcher


class UnitTests(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        if os.path.isfile("1_test.txt"):
            os.remove("1_test.txt")
        if os.path.isfile("2_test.txt"):
            os.remove("2_test.txt")

    def test_search_and_download_namibsun(self):

        searcher = PackSearcher(["namibsun"])
        packs = searcher.search("1_test.txt") + searcher.search("2_test.txt")

        XDCCDownloader(packs[0].get_server(), "random").download(packs)

        self.assertTrue(os.path.isfile("1_test.txt"))
        self.assertTrue(os.path.isfile("2_test.txt"))



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
from typing import List
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.pack_searchers.procedures.nibl import find_nibl_packs
from xdcc_dl.pack_searchers.procedures.ixirc import find_ixirc_packs
from xdcc_dl.pack_searchers.procedures.namibsun import find_namibsun_packs
from xdcc_dl.pack_searchers.procedures.horriblesubs import find_horriblesubs_packs
from xdcc_dl.pack_searchers.procedures.intel_haruhichan import find_intel_haruhichan_packs


class PackSearcher(object):
    """
    Class that offers various methods to search different XDCC Packlists for XDCC Packs
    """

    procedure_map = {"nibl": find_nibl_packs,
                     "ixirc": find_ixirc_packs,
                     "namibsun": find_namibsun_packs,
                     "horriblesubs": find_horriblesubs_packs,
                     "intel_haruhichan": find_intel_haruhichan_packs}

    @staticmethod
    def get_available_pack_searchers() -> List[str]:
        """
        Returns a list of available pack searchers by name

        :return: A list of the names of the pack searchers
        """
        return list(PackSearcher.procedure_map.keys())

    def __init__(self, procedures: List[str] = list(procedure_map.keys())) -> None:
        """
        Initializes the Packsearcher with a list of procedures to consider
        All procedures are used by default

        :raises:           KeyError, if an invali procedure was specified
        :param procedures: List of procedures to use during the search
        """
        self.procedures = []
        for procedure in procedures:
            self.procedures.append(self.procedure_map[procedure])

    def search(self, search_phrase: str) -> List[XDCCPack]:
        """
        Conducts the actual search for the XDCC packs

        :param search_phrase: The search phrase to use while conducting the search
        :return:              A list of search results as XDCCPack objects
        """
        results = []
        for procedure in self.procedures:
            results += procedure(search_phrase)

        for result in results:
            result.set_original_filename(result.get_filename())

        return results



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""


class FGColors:
    """
    Foreground Colours for colouring console output
    """

    DEFAULT = '\033[39m'
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    LIGHT_GRAY = '\033[37m'
    DARK_GRAY = '\033[90m'
    LIGHT_RED = '\033[91m'
    LIGHT_GREEN = '\033[92m'
    LIGHT_YELLOW = '\033[93m'
    LIGHT_BLUE = '\033[94m'
    LIGHT_MAGENTA = '\033[95m'
    LIGHT_CYAN = '\033[96m'
    WHITE = '\033[97m'


class BGColors:
    """
    Background Colours for colouring console output
    """

    DEFAULT = '\033[49m'
    BLACK = '\033[40m'  # WHITE
    RED = '\033[41m'
    GREEN = '\033[42m'  # YELLOW
    YELLOW = '\033[43m'
    BLUE = '\033[44m'
    MAGENTA = '\033[45m'
    CYAN = '\033[46m'
    LIGHT_GRAY = '\033[47m'
    DARK_GRAY = '\033[100m'
    LIGHT_RED = '\033[101m'
    LIGHT_GREEN = '\033[102m'
    LIGHT_YELLOW = '\033[103m'
    LIGHT_BLUE = '\033[104m'
    LIGHT_MAGENTA = '\033[105m'
    LIGHT_CYAN = '\033[106m'
    WHITE = '\033[107m'


class LoggingTypes:
    """
    The different logging types.

    Each logging type is assigned a background and foreground colour and a priority. The priority
    determines if the type of logging element is shown at all. In General, the logging levels mean:

        0: Must be shown under all circumstances except pure GUI output
        1: Will definitely be shown in a CLI environment
        2: Will generally be shown in a CLI environment
        3: Will be shown in environments where detailed output is desired
        4: Will generally not be shown, may lead to excessive amount of output
        5: Shows all undefined events. WILL lead to excessive output
        6: Has nothing to do with the task at hand

    Colour Coding:

        YELLOW/L_YELLOW BG:                    DCC / Downloads
        CYAN BG:                               WHOIS
        GREEN BG:                              CHANNELS
        BLUE BG:                               Messages
        DEFAULT BG + GREY/L_GREY/WHITE FG:     Welcome Messages, Message of the day, CTCP Version
        DEFAULT BG + DEFAULT/L_GREEN/L_RED FG: Connection
        DEFAULT BG + BLUE/L_BLUE FG:           Private Message/Notice
        DEFAULT BG + YELLOW/L_YELLOW FG:       Public Message/Notice
        DEFAULT BG + MANGENTA/L_MAGENTA FG:    Undefined events
    """

    INVISIBLE            = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.DEFAULT,        "priority": 0}
    DEFAULT              = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.DEFAULT,        "priority": 1}

    CONNECTION_ATTEMPT   = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.DEFAULT,        "priority": 2}
    CONNECTION_SUCCESS   = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.LIGHT_GREEN,    "priority": 2}
    CONNECTION_FAILURE   = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.LIGHT_RED,      "priority": 2}
    BANNED               = {"bg_color": BGColors.LIGHT_RED,    "fg_color": FGColors.LIGHT_MAGENTA,  "priority": 1}

    WHOIS_SEND           = {"bg_color": BGColors.CYAN,         "fg_color": FGColors.BLACK,          "priority": 2}
    WHOIS_SUCCESS        = {"bg_color": BGColors.CYAN,         "fg_color": FGColors.WHITE,          "priority": 2}
    WHOIS_NO_RESULT      = {"bg_color": BGColors.CYAN,         "fg_color": FGColors.LIGHT_RED,      "priority": 2}
    WHOIS_USER           = {"bg_color": BGColors.CYAN,         "fg_color": FGColors.DARK_GRAY,      "priority": 4}
    WHOIS_SERVER         = {"bg_color": BGColors.CYAN,         "fg_color": FGColors.LIGHT_GRAY,     "priority": 4}

    CHANNEL_JOIN_ATTEMPT = {"bg_color": BGColors.GREEN,        "fg_color": FGColors.BLACK,          "priority": 2}
    CHANNEL_JOIN_SUCCESS = {"bg_color": BGColors.GREEN,        "fg_color": FGColors.LIGHT_GREEN,    "priority": 2}
    CHANNEL_USERS        = {"bg_color": BGColors.GREEN,        "fg_color": FGColors.LIGHT_BLUE,     "priority": 6}
    CHANNEL_TOPIC        = {"bg_color": BGColors.GREEN,        "fg_color": FGColors.MAGENTA,        "priority": 6}
    CHANNEL_QUIT         = {"bg_color": BGColors.GREEN,        "fg_color": FGColors.RED,            "priority": 6}
    CHANNEL_PART         = {"bg_color": BGColors.GREEN,        "fg_color": FGColors.LIGHT_MAGENTA,  "priority": 6}
    CHANNEL_KICK         = {"bg_color": BGColors.GREEN,        "fg_color": FGColors.LIGHT_RED,      "priority": 6}
    CHANNEL_MODE_CHANGE  = {"bg_color": BGColors.GREEN,        "fg_color": FGColors.LIGHT_YELLOW,   "priority": 6}
    CHANNEL_ACTION       = {"bg_color": BGColors.GREEN,        "fg_color": FGColors.YELLOW,         "priority": 6}
    CHANNEL_NICK_CHANGED = {"bg_color": BGColors.GREEN,        "fg_color": FGColors.LIGHT_CYAN,     "priority": 6}

    MESSAGE_SEND         = {"bg_color": BGColors.BLUE,          "fg_color": FGColors.BLACK,         "priority": 1}
    MESSAGE_RETRY        = {"bg_color": BGColors.BLUE,          "fg_color": FGColors.WHITE,         "priority": 1}

    ALREADY_REQUESTED    = {"bg_color": BGColors.BLUE,          "fg_color": FGColors.LIGHT_RED,     "priority": 2}

    INCORRECT_FILE       = {"bg_color": BGColors.YELLOW,        "fg_color": FGColors.LIGHT_MAGENTA, "priority": 1}
    DCC_SEND_HANDSHAKE   = {"bg_color": BGColors.YELLOW,        "fg_color": FGColors.BLACK,         "priority": 2}
    DCC_RESUME_REQUEST   = {"bg_color": BGColors.YELLOW,        "fg_color": FGColors.DARK_GRAY,     "priority": 2}
    DCC_RESUME_SUCCESS   = {"bg_color": BGColors.YELLOW,        "fg_color": FGColors.WHITE,         "priority": 2}
    DCC_RESUME_FAILED    = {"bg_color": BGColors.YELLOW,        "fg_color": FGColors.LIGHT_RED,     "priority": 2}

    DOWNLOAD_START       = {"bg_color": BGColors.LIGHT_YELLOW,  "fg_color": FGColors.LIGHT_GREEN,   "priority": 1}
    DOWNLOAD_RESUME      = {"bg_color": BGColors.LIGHT_YELLOW,  "fg_color": FGColors.LIGHT_BLUE,    "priority": 1}
    DOWNLOAD_PROGRESS    = {"bg_color": BGColors.LIGHT_YELLOW,  "fg_color": FGColors.BLACK,         "priority": 1}
    DOWNLOAD_WAS_DONE    = {"bg_color": BGColors.LIGHT_YELLOW,  "fg_color": FGColors.RED,           "priority": 1}
    DOWNLOAD_INCOMPLETE  = {"bg_color": BGColors.LIGHT_YELLOW,  "fg_color": FGColors.LIGHT_RED,     "priority": 2}

    PRIVATE_NOTICE       = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.BLUE,           "priority": 3}
    PRIVATE_MESSAGE      = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.LIGHT_BLUE,     "priority": 3}
    PUBLIC_NOTICE        = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.YELLOW,         "priority": 4}
    PUBLIC_MESSAGE       = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.LIGHT_YELLOW,   "priority": 4}

    MESSAGE_OF_THE_DAY   = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.LIGHT_GRAY,     "priority": 4}
    WELCOME              = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.DARK_GRAY,      "priority": 4}
    CTCP_VERSION         = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.WHITE,          "priority": 4}
    PING                 = {"bg_color": BGColors.LIGHT_GREEN,  "fg_color": FGColors.BLACK,          "priority": 4}

    EVENT                = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.MAGENTA,        "priority": 5}
    EVENT_TEXT           = {"bg_color": BGColors.DEFAULT,      "fg_color": FGColors.LIGHT_MAGENTA,  "priority": 5}



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import unittest
from xdcc_dl.entities.IrcServer import IrcServer


class UnitTests(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_server_entity_constructor(self):

        server = IrcServer("irc.namibsun.net")
        self.assertEqual(server.get_address(), "irc.namibsun.net")
        self.assertEqual(server.get_port(), 6667)

    def test_server_entity_constructor_with_port(self):

        server = IrcServer("irc.namibsun.net", 9000)
        self.assertEqual(server.get_address(), "irc.namibsun.net")
        self.assertEqual(server.get_port(), 9000)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import irc.client
import jaraco.logging
from jaraco.stream import buffer
from xdcc_dl.entities.User import User
from xdcc_dl.logging.Logger import Logger
from xdcc_dl.entities.IrcServer import IrcServer
from xdcc_dl.xdcc.layers.helpers.Variables import Variables
# noinspection PyPep8Naming
from xdcc_dl.logging.LoggingTypes import LoggingTypes as LOG
from xdcc_dl.xdcc.layers.helpers.ConnectionStates import ConnectionStates


class IgnoreErrorsBuffer(buffer.DecodingLineBuffer):
    """
    Decoding Buffer for the IRC Client that ignores any errors, like UTF8
    decoding errors, which may appear with some messages.
    """
    def handle_exception(self) -> None:
        """
        Handles the Exception itself, does nothing.
        :return: None
        """
        pass  # pragma: no cover


class Disconnect(Exception):
    """
    Class that gets raised when a normal disconnect occurs
    """
    pass


class NetworkError(Exception):
    """
    Class that gets raised when a network error occurs
    """
    pass


class Banned(Exception):
    """
    Exception that gets raised when a network error occurs
    """
    pass


class BaseIrclient(irc.client.SimpleIRCClient, ConnectionStates, Variables):
    """
    The Base IRC Client that defines the necessary features that an IRC Client must be able to do.
    Layer 0 of the XDCC Bot
    """

    def __init__(self, server: IrcServer or str, user: User or str, logger: Logger or int = 0):
        """
        Initializes the Client's Server Connection Information and disables Buffer Errors
        The parameters can all be initialized with either a string/int representing the object's
        main value or the classes themselves

        :param server: The IRC Server to which th client will attempt to connect to
                       If a string was provided, create IrcServer with default ort 6667
        :param user:   The User to log in to the IRC Server with
                       If a string was provided, create user object with that username
        :param logger: The logger used to print informational messages to the console
                       If an int was provided, creates a standard console logger with the specified verbosity level
        """
        super().__init__()
        Variables.__init__(self)
        ConnectionStates.__init__(self)

        jaraco.logging.log_level("0")

        irc.client.ServerConnection.buffer_class = IgnoreErrorsBuffer
        irc.client.SimpleIRCClient.buffer_class = IgnoreErrorsBuffer

        self.user = user if user.__class__ == User else User(user)
        self.server = server if server.__class__ == IrcServer else IrcServer(server)
        self.logger = logger if logger.__class__ == Logger else Logger(logger)

    def connect(self) -> None:
        """
        Connects the IRC Client to the IRC Server

        :raises: NetworkError if the connection to the server did not succeed
        :return: None
        """
        self.logger.log("Connecting to server:  " + self.server.get_address(), LOG.CONNECTION_ATTEMPT)
        self.logger.log("Using Port:            " + str(self.server.get_port()), LOG.CONNECTION_ATTEMPT)
        self.logger.log("As User:               " + self.user.get_name(), LOG.CONNECTION_ATTEMPT)

        super().connect(self.server.get_address(), int(self.server.get_port()), self.user.get_name())
        self.logger.log("Established Connection to Server", LOG.CONNECTION_SUCCESS)
        self.connected_to_server = True

    def start(self) -> None:
        """
        Starts the IRC Connection

        :raises: NetworkError if the connection to the server did not succeed
        :return: None
        """
        network_error = ""

        try:
            self.connect()
            super().start()
        except irc.client.ServerConnectionError:
            self.logger.log("Failed to connect to Server", LOG.CONNECTION_FAILURE)
            network_error = "Failed to connect to Server"
        except Banned:
            self.logger.log("Failed to connect due to a ban", LOG.BANNED)
            network_error = "Failed to connect due to a ban"
        except Exception as e:
            try:
                self.quit()
            except (Disconnect, irc.client.ServerNotConnectedError):
                pass
            if str(type(e)) != "<class 'xdcc_dl.xdcc.layers.irc.BaseIrcClient.Disconnect'>":
                self.connected_to_server = False
                raise e

        try:
            self.quit()
        except Disconnect:
            pass

        if network_error:
            raise NetworkError(network_error)

    def quit(self) -> None:
        """
        Forcibly closes the connection

        :return: None
        """

        try:
            self.connection.disconnect()
            self.connection.quit()
        except (Disconnect, irc.client.ServerNotConnectedError):
            try:
                self.connection.quit()
            except (Disconnect, irc.client.ServerNotConnectedError):
                pass

        if self.connected_to_server:
            self.connected_to_server = False
            raise Disconnect()

    # noinspection PyMethodMayBeStatic
    def on_disconnect(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Method called whenever the IRC connection is disconnected

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :raises:           Disconnect, when the connection was disconnected by non-fatal means
        :return:           None
        """
        raise Disconnect()

    # noinspection PyMethodMayBeStatic
    def on_error(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Method called whenever the IRC connection throws an error event, which means that the user is banned

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :raises:           Disconnect, when the connection was disconnected by non-fatal means
        :return:           None
        """
        raise Banned()



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import cfscrape
from typing import List
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.IrcServer import IrcServer


def find_horriblesubs_packs(search_phrase: str) -> List[XDCCPack]:
    """
    Method that conducts the xdcc pack search for xdcc.horriblesubs.info

    :return: the search results as a list of XDCCPack objects
    """
    if not search_phrase:
        return []

    search_query = search_phrase.replace(" ", "%20")
    search_query = search_query.replace("!", "%21")

    url = "http://xdcc.horriblesubs.info/search.php?t=" + search_query
    scraper = cfscrape.create_scraper()
    results = scraper.get(url).text.split(";")

    packs = []
    for result in results:

        try:
            botname = get_attribute(result, "b")
            filename = get_attribute(result, "f")
            filesize = get_attribute(result, "s")
            packnumber = get_attribute(result, "n")
            pack = XDCCPack(IrcServer("irc.rizon.net"), botname, packnumber)
            pack.set_filename(filename)
            pack.set_size(filesize)
            packs.append(pack)

        except IndexError:  # In case the line is not parseable
            pass
    return packs


def get_attribute(pack_string: str, attribute: str) -> str or int:
    """
    Parses a horriblesubs pack string for an attribute

    :param pack_string: the string to parse
    :param attribute:   the attribute to get
    :return:            the requested attribute
    """
    attribute_value = pack_string.split(attribute + ":")[1].split(",")[0]
    if attribute in ["b", "f"]:  # bot or file is encased in quotes
        return attribute_value.split("\"")[1].split("\"")[0]
    elif attribute == "s":
        return attribute_value + "M"  # Size is in Megabyte
    elif attribute == "n":
        return int(attribute_value)
    else:  # pragma: no cover
        pass



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
try:
    from PyQt5.QtCore import Qt
    from PyQt5.QtTest import QTest
    from xdcc_dl.gui.XDCCDownloaderGui import XDCCDownloaderGui, QApplication
except ImportError:
    Qt = QTest = XDCCDownloaderGui = QApplication = None

import os
import sys
import time
import unittest
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.IrcServer import IrcServer


class UnitTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        if QApplication is None or sys.version_info[0] < 3:
            raise unittest.SkipTest("Skipping on python 2 or import error")

        sys.argv = [sys.argv[0], "-platform", "minimal"]
        cls.app = QApplication(sys.argv)

    def setUp(self):
        sys.argv = [sys.argv[0], "-platform", "minimal"]
        self.form = XDCCDownloaderGui()

    def tearDown(self):
        self.form.downloading = False
        self.form.searching = False
        self.form.destroy()
        if os.path.isfile("1_test.txt"):
            os.remove("1_test.txt")
        if os.path.isfile("2_test.txt"):
            os.remove("2_test.txt")
        if os.path.isfile("3_test.txt"):
            os.remove("3_test.txt")

    def test_default_values(self):
        self.assertEqual(os.getcwd(), self.form.destination_edit.text())
        self.assertEqual("irc.rizon.net", self.form.server_edit.text())

    def test_searching(self):
        self.form.search_term_edit.setText("1_test.txt")
        self.assertEqual(self.form.search_term_edit.text(), "1_test.txt")
        self.form.search_engine_combo_box.setCurrentText("namibsun")
        self.assertEqual(self.form.search_engine_combo_box.currentText(), "namibsun")
        QTest.mouseClick(self.form.search_button, Qt.LeftButton)

        while self.form.searching:
            pass

        self.form.refresh_search_results()
        self.form.search_result_tree_widget.selectAll()

        self.assertEqual(1, len(self.form.search_results))
        self.assertEqual(self.form.search_results[0].get_bot(), "xdcc_servbot")
        self.assertEqual(1, self.form.search_result_tree_widget.invisibleRootItem().childCount())

    def test_adding_search_result(self):
        self.test_searching()

        QTest.mouseClick(self.form.rigth_arrow_button, Qt.LeftButton)
        self.assertEqual(1, len(self.form.download_queue))
        self.assertEqual(1, self.form.download_queue_list_widget.count())

    def test_removing_pack_from_queue(self):
        self.form.download_queue = [XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 1)]
        self.form.refresh_download_queue()
        self.assertEqual(1, self.form.download_queue_list_widget.count())

        self.form.download_queue_list_widget.selectAll()
        QTest.mouseClick(self.form.left_arrow_button, Qt.LeftButton)

        self.assertEqual(0, len(self.form.download_queue))
        self.assertEqual(0, self.form.download_queue_list_widget.count())

    def test_adding_manual_pack(self):
        self.form.message_edit.setText("/msg xdcc_servbot xdcc send #1")
        self.form.server_edit.setText("irc.namibsun.net")
        self.assertEqual(self.form.message_edit.text(), "/msg xdcc_servbot xdcc send #1")
        self.assertEqual(self.form.server_edit.text(), "irc.namibsun.net")

        QTest.mouseClick(self.form.add_button, Qt.LeftButton)
        self.form.refresh_download_queue()
        self.assertEqual(1, len(self.form.download_queue))
        self.assertEqual(1, self.form.download_queue_list_widget.count())

    def test_download_packs(self):
        self.form.download_queue = [XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 1),
                                    XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 2),
                                    XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 3)]
        self.form.refresh_download_queue()

        self.assertEqual(self.form.destination_edit.text(), os.getcwd())
        QTest.mouseClick(self.form.download_button, Qt.LeftButton)

        while self.form.downloading:
            pass

        self.assertTrue(os.path.isfile("1_test.txt"))
        self.assertTrue(os.path.isfile("2_test.txt"))
        self.assertTrue(os.path.isfile("3_test.txt"))
        self.form.show_download_complete_message_signal.emit("")

    def test_download_while_downloading(self):
        self.form.download_queue = [XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 1),
                                    XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 3)]
        self.form.refresh_download_queue()
        QTest.mouseClick(self.form.download_button, Qt.LeftButton)

        time.sleep(0.5)

        self.form.download_queue = [XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 2)]
        self.form.refresh_download_queue()
        QTest.mouseClick(self.form.download_button, Qt.LeftButton)

        while self.form.downloading:
            pass

        self.assertTrue(os.path.isfile("1_test.txt"))
        self.assertFalse(os.path.isfile("2_test.txt"))
        self.assertTrue(os.path.isfile("3_test.txt"))
        self.form.show_download_complete_message_signal.emit("")

    def test_search_while_searching(self):

        self.form.search_term_edit.setText("1_test.txt")
        self.assertEqual(self.form.search_term_edit.text(), "1_test.txt")
        self.form.search_engine_combo_box.setCurrentText("All")
        self.assertEqual(self.form.search_engine_combo_box.currentText(), "All")

        QTest.mouseClick(self.form.search_button, Qt.LeftButton)

        self.form.search_term_edit.setText("2_test.txt")
        self.assertEqual(self.form.search_term_edit.text(), "2_test.txt")
        self.form.search_engine_combo_box.setCurrentText("namibsun")
        self.assertEqual(self.form.search_engine_combo_box.currentText(), "namibsun")

        QTest.mouseClick(self.form.search_button, Qt.LeftButton)

        while self.form.searching:
            pass

        self.form.refresh_search_results()
        self.form.search_result_tree_widget.selectAll()

        self.assertEqual(1, len(self.form.search_results))
        self.assertEqual(self.form.search_results[0].get_bot(), "xdcc_servbot")
        self.assertEqual(1, self.form.search_result_tree_widget.invisibleRootItem().childCount())

    def test_spinner(self):
        self.form.downloading = True
        self.form.spinner_start_signal.emit("download")
        self.form.searching = True
        self.form.spinner_start_signal.emit("search")
        time.sleep(1)
        self.form.downloading = False
        self.form.searching = False
        time.sleep(1)

    def test_spin_update(self):
        self.assertEqual(self.form.download_button.text(), "Download")
        self.form.spinner_updater_signal.emit(self.form.download_button, "NotDownload")
        self.assertEqual(self.form.download_button.text(), "NotDownload")

    def test_invalid_download_directory(self):
        self.form.destination_edit.setText("")
        QTest.mouseClick(self.form.download_button, Qt.LeftButton)

    def test_move_selection_up_or_down(self):
        self.form.download_queue = [XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 1),
                                    XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 2),
                                    XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 3)]
        self.form.refresh_download_queue()

        self.form.download_queue_list_widget.selectAll()
        QTest.mouseClick(self.form.up_arrow_button, Qt.LeftButton)

        self.form.download_queue_list_widget.selectAll()
        QTest.mouseClick(self.form.down_arrow_button, Qt.LeftButton)

        self.assertEqual(self.form.download_queue_list_widget.count(), 3)

    def test_progress_update(self):
        self.assertEqual(self.form.single_progress_bar.value(), 0.0)
        self.assertEqual(self.form.total_progress_bar.value(), 0.0)
        self.form.progress_update_signal.emit(50.0, 75.0)
        self.assertEqual(self.form.single_progress_bar.value(), 50.0)
        self.assertEqual(self.form.total_progress_bar.value(), 75.0)

    def test_removing_queue_items(self):
        self.form.download_queue = [XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 1),
                                    XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 2),
                                    XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 3)]
        self.form.refresh_download_queue()
        self.form.download_queue_list_widget.selectAll()

        self.assertEqual(self.form.download_queue_list_widget.count(), 3)
        self.assertEqual(len(self.form.download_queue), 3)
        self.assertEqual(len(self.form.download_queue_list_widget.selectedIndexes()), 3)

        QTest.mouseClick(self.form.left_arrow_button, Qt.LeftButton)

        self.assertEqual(self.form.download_queue_list_widget.count(), 0)
        self.assertEqual(len(self.form.download_queue), 0)
        self.assertEqual(len(self.form.download_queue_list_widget.selectedIndexes()), 0)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import unittest
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.IrcServer import IrcServer
from xdcc_dl.xdcc.layers.helpers.BotChannelMapper import BotChannelMapper
from xdcc_dl.xdcc.layers.irc.BotFinder import BotFinder, BotNotFoundException


class TestException(Exception):
    pass


class UnitTests(unittest.TestCase):

    def setUp(self):
        self.client = BotFinder

    def tearDown(self):
        self.client.quit()

    def test_whois_query(self):

        class Tester(BotFinder):

            joined_correct_channel = False
            all_whois_counter = 0

            def on_welcome(self, conn, event):
                self.connection.whois("xdcc_servbot")

            # noinspection PyUnusedLocal
            def on_join(self, conn, event):
                if event.source.startswith(self.user.get_name()):
                    self.joined_correct_channel = event.target.lower() == "#bots"

                    if self.all_whois_counter == 3:
                        raise TestException()

            def on_whoischannels(self, conn, event):
                self.all_whois_counter += 1
                super().on_whoischannels(conn, event)

            def on_whoisserver(self, conn, event):
                self.all_whois_counter += 1
                super().on_whoisserver(conn, event)

            def on_endofwhois(self, conn, event):
                self.all_whois_counter += 1
                super().on_endofwhois(conn, event)

                if self.joined_correct_channel:
                    raise TestException()

        self.client = Tester("irc.namibsun.net", "random")

        try:
            self.client.start()
            self.assertTrue(False)
        except TestException:
            self.assertTrue(self.client.joined_correct_channel)
            self.assertEqual(self.client.all_whois_counter, 3)

    def test_invalid_whois_query(self):

        class Tester(BotFinder):

            incorrect_nick_detected = False

            def on_welcome(self, conn, event):
                self.current_pack = XDCCPack(IrcServer("irc.namibsun.net"), "notexistingbot", 1)
                self.connection.whois("notexistingbot")

            # noinspection PyUnusedLocal
            def on_nosuchnick(self, conn, event):
                try:
                    super().on_nosuchnick(conn, event)
                except BotNotFoundException:
                    self.incorrect_nick_detected = True
                raise TestException()

        self.client = Tester("irc.namibsun.net", "random")

        try:
            self.client.start()
            self.assertTrue(False)
        except TestException:
            self.assertTrue(self.client.incorrect_nick_detected)

    def test_stray_whois(self):

        class Tester(BotFinder):

            stray_whois_detected = False

            def on_welcome(self, conn, event):
                self.current_pack = XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 1)
                self.connection.whois("notexistingbot")

            # noinspection PyUnusedLocal
            def on_nosuchnick(self, conn, event):
                super().on_nosuchnick(conn, event)
                self.stray_whois_detected = True
                raise TestException()

        self.client = Tester("irc.namibsun.net", "random")

        try:
            self.client.start()
            self.assertTrue(False)
        except TestException:
            self.assertTrue(self.client.stray_whois_detected)

    def test_channel_less_whois_query(self):

        class Tester(BotFinder):

            dummy_channel_joined = False

            def on_welcome(self, conn, event):
                self.current_pack = XDCCPack(self.server, "channelless_bot", 1)
                self.connection.whois(self.current_pack.get_bot())

            # noinspection PyUnusedLocal
            def on_nosuchnick(self, conn, event):
                self.on_endofwhois(conn, event)

            # noinspection PyUnusedLocal
            def on_join(self, conn, event):
                if event.source == self.user.get_name():
                    self.dummy_channel_joined = True
                    raise TestException()

        self.client = Tester("irc.namibsun.net", "random")

        try:
            self.client.start()
            self.assertTrue(False)
        except TestException:
            self.assertTrue(self.client.dummy_channel_joined)

    def test_mapped_channel(self):

        class Tester(BotFinder):

            correct_channel_joined = False

            def on_welcome(self, conn, event):
                self.current_pack = XDCCPack(self.server, "HelloKitty", 1)
                self.connection.whois(self.current_pack.get_bot())

            # noinspection PyUnusedLocal
            def on_nosuchnick(self, conn, event):
                self.on_endofwhois(conn, event)

            # noinspection PyUnusedLocal
            def on_join(self, conn, event):
                if event.target.lower() == BotChannelMapper.map(self.current_pack.get_bot()).lower():
                    self.correct_channel_joined = True
                    raise TestException()

        self.client = Tester("irc.rizon.net", "random")

        try:
            self.client.start()
            self.assertTrue(False)
        except TestException:
            self.assertTrue(self.client.correct_channel_joined)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import unittest
from xdcc_dl.entities.Progress import Progress
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.xdcc.layers.xdcc.DownloadHandler import DownloadHandler


class UnitTests(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        if os.path.isfile("testfile.mkv"):
            os.remove("testfile.mkv")
        if os.path.isfile("1_test.txt"):
            os.remove("1_test.txt")
        if os.path.isfile("2_test.txt"):
            os.remove("2_test.txt")
        if os.path.isfile("3_test.txt"):
            os.remove("3_test.txt")

    def test_incomplete_download(self):

        class Tester(DownloadHandler):

            download_incomplete_check = False
            download_completed_check = True

            def on_welcome(self, conn, event):

                self.progress = Progress(1)

                self.file = open("testfile.mkv", 'wb')
                self.file.write(b"test")
                self.filesize = 10000000

                self.current_pack = XDCCPack(self.server, "xdcc_servbot", 1)
                self.pack_queue = []

                self.logger.log = self.logging_handler

                self.on_dcc_disconnect(conn, event)

            # noinspection PyUnusedLocal
            def logging_handler(self, string, formatting, carriage_return=None):
                if string == "Download Incomplete, Trying again.":
                    self.download_incomplete_check = True
                elif "Download completed in" in string:
                    self.download_completed_check = True
                else:
                    print(string)

        client = Tester("irc.namibsun.net", "random")
        client.start()

        self.assertTrue(os.path.isfile("1_test.txt"))
        self.assertTrue(client.download_completed_check)
        self.assertTrue(client.download_incomplete_check)

    # XDCCInitator-related tests

    def test_extended_privnotice(self):

        class Tester(DownloadHandler):

            def on_welcome(self, conn, event):

                self.current_pack = XDCCPack(self.server, "xdcc_servbot", 3)
                self.pack_queue = []
                self.progress = Progress(1)

                event.arguments = []
                self.on_privnotice(conn, event)
                event.arguments = ["You will have to re-send that, to the bot that transferred the file."]
                self.on_privnotice(conn, event)

        Tester("irc.namibsun.net", "random").start()
        self.assertTrue(os.path.isfile("3_test.txt"))

    def test_failed_dcc_resume_detection(self):

        class Tester(DownloadHandler):

            def on_welcome(self, conn, event):
                self.current_pack = XDCCPack(self.server, "xdcc_servbot", 2)
                self.pack_queue = []
                self.progress = Progress(1)

                with open("2_test.txt", 'w') as f:
                    f.write("Test")

                self.dcc_resume_requested = True
                self.connection.whois(self.current_pack.get_bot())

        Tester("irc.namibsun.net", "random").start()
        self.assertTrue(os.path.isfile("2_test.txt"))



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import unittest
from threading import Thread
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.Progress import Progress
from xdcc_dl.entities.IrcServer import IrcServer
from xdcc_dl.xdcc.MultipleServerDownloader import MultipleServerDownloader


class UnitTests(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        if os.path.isfile("1_test.txt"):
            os.remove("1_test.txt")
        if os.path.isfile("2_test.txt"):
            os.remove("2_test.txt")
        if os.path.isfile("3_test.txt"):
            os.remove("3_test.txt")

    def test_download_multiple_packs_same_server(self):

        progress = Progress(2)

        downloader = MultipleServerDownloader("random")

        downloader.download([XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 1),
                             XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 3)], progress)

        self.assertTrue(os.path.isfile("1_test.txt"))
        self.assertTrue(os.path.isfile("3_test.txt"))

        self.assertEqual(progress.get_single_progress_percentage(), 100.0)
        self.assertEqual(progress.get_total_percentage(), 100.0)

    def test_download_multiple_packs_different_servers(self):

        progress = Progress(2)

        downloader = MultipleServerDownloader("random")

        downloader.download([XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 2),
                             XDCCPack(IrcServer("namibsun.net"), "xdcc_servbot", 3)], progress)

        self.assertTrue(os.path.isfile("2_test.txt"))
        self.assertTrue(os.path.isfile("3_test.txt"))

        self.assertEqual(progress.get_single_progress_percentage(), 100.0)
        self.assertEqual(progress.get_total_percentage(), 100.0)

    def test_quitting(self):

        packs = [XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 1)]

        downloader = MultipleServerDownloader("random")
        downloader.quit()
        downloader.download(packs)
        downloader.quit()

        self.assertFalse(os.path.isfile("1_test.txt"))



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import unittest
from xdcc_dl.pack_searchers.procedures.horriblesubs import find_horriblesubs_packs


class UnitTests(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_empty_query(self):
        packs = find_horriblesubs_packs("")
        self.assertEqual(len(packs), 0)

    def test_gin_txt(self):
        packs = find_horriblesubs_packs("Gin.txt")
        self.assertEqual(len(packs), 1)

        pack = packs[0]
        self.assertEqual(pack.get_packnumber(), 1)
        self.assertEqual(pack.get_server().get_address(), "irc.rizon.net")
        self.assertEqual(pack.get_filepath(), os.path.join(os.getcwd(), pack.get_filename()))
        self.assertEqual(pack.get_filename(), "Gin.txt")
        self.assertEqual(pack.get_bot(), "Ginpachi-Sensei")

    def test_larger_result(self):
        packs = find_horriblesubs_packs("One Punch Man")
        self.assertLess(10, len(packs))

    def test_non_result_query(self):
        self.assertEqual(0, len(find_horriblesubs_packs("sdgyfdhkdashsahdqhdsadlsajdhsaohdsausahoashdsahdlahdsah")))



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import sys
import argparse
from xdcc_dl.metadata import SentryLogger
from xdcc_dl.xdcc.XDCCDownloader import XDCCDownloader
from xdcc_dl.tui.XDCCDownloaderTui import XDCCDownloaderTui
from xdcc_dl.entities.XDCCPack import xdcc_packs_from_xdcc_message

try:
    from xdcc_dl.gui.XDCCDownloaderGui import start as start_gui
except ImportError:  # pragma: no cover
    start_gui = None


def main() -> None:
    """
    Starts the main method of the program

    :return: None
    """
    try:

        parser = argparse.ArgumentParser()
        parser.add_argument("message", nargs='?',
                            help="An XDCC Message. Supports ranges (1-100) and also ranges with steps (1-100,2)")
        parser.add_argument("-s", "--server",
                            help="Specifies the IRC Server. Defaults to irc.rizon.net")
        parser.add_argument("-d", "--destination",
                            help="Specifies the target download destination. Defaults to " + os.getcwd())
        parser.add_argument("-u", "--username",
                            help="Specifies the username")
        parser.add_argument("-v", "--verbosity", type=int, default=1,
                            help="Specifies the verbosity of the output on a scale of 1-7. Default: 1")
        parser.add_argument("-g", "--gui", action="store_true",
                            help="Starts the XDCC Downloader GUI")
        parser.add_argument("-t", "--tui", action="store_true",
                            help="Starts the XDCC Downloader TUI")
        args = parser.parse_args()

        if args.message:

            destination = os.getcwd() if not args.destination else args.destination
            server = "irc.rizon.net" if not args.server else args.server
            user = "random" if not args.username else args.username
            verbosity = args.verbosity

            packs = xdcc_packs_from_xdcc_message(args.message, destination, server)
            downloader = XDCCDownloader(server, user, verbosity)
            results = downloader.download(packs)
            downloader.quit()

            max_length = max(map(lambda x: len(x.get_filepath()), results.keys()))
            for result in results:
                print(result.get_filepath().ljust(max_length) + " - " + results[result])

        elif args.gui:  # pragma: no cover
            if start_gui is not None:
                start_gui()
            else:
                print("Error: PyQt5 not installed")

        elif args.tui:  # pragma: no cover
            XDCCDownloaderTui().start()

        else:
            print("No arguments passed. See --help for more details")
            sys.exit(0)

    except KeyboardInterrupt:
        print("Thanks for using xdcc-downloader!")
    except Exception as e:  # pragma: no cover
        SentryLogger.sentry.captureException()
        raise e


if __name__ == "__main__":  # pragma: no cover

    if sys.platform == "win32" and len(sys.argv) == 1:
        sys.argv.append("-g")
    main()



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import time
import unittest
from xdcc_dl.entities.Progress import Progress


class UnitTests(unittest.TestCase):

    def setUp(self):
        self.progress = Progress(10)
        self.callback_called = False

    def tearDown(self):
        pass

    def test_single_progress(self):

        self.progress.set_single_progress(100)
        self.progress.set_single_progress_total(1000)
        self.assertEqual(self.progress.get_single_progress(), 100)
        self.assertEqual(self.progress.get_single_progress_total(), 1000)
        self.assertEqual(self.progress.get_single_progress_percentage(), 10.0)

        self.progress.add_single_progress(100)
        self.assertEqual(self.progress.get_single_progress(), 200)
        self.assertEqual(self.progress.get_single_progress_percentage(), 20.0)

        self.assertLess(self.progress.get_total_percentage(), 10.0)
        self.assertLess(0.0, self.progress.get_total_percentage())

    def test_next_file(self):

        self.assertEqual(self.progress.get_total_percentage(), 0.0)
        self.progress.next_file()
        self.assertEqual(self.progress.get_total_percentage(), 10.0)

    def test_callback_handler(self):

        def callback_tester(single_progress, single_total, single_progress_percentage,
                            total_progress, total_total, total_percentage, current_speed, average_speed):

            self.callback_called = True
            self.assertEqual(single_progress, 300)
            self.assertEqual(single_total, 1000)
            self.assertEqual(single_progress_percentage, 30.0)
            self.assertEqual(total_progress, 0)
            self.assertEqual(total_total, 2)
            self.assertLess(total_percentage, 50.0)
            self.assertLess(0.0, total_percentage)
            self.assertAlmostEqual(current_speed, 100, delta=5)
            self.assertAlmostEqual(average_speed, 100, delta=5)

        progress_with_callback = Progress(2, callback=callback_tester)
        progress_with_callback.set_single_progress_total(1000)
        time.sleep(3)
        progress_with_callback.add_single_progress(300)

        self.assertTrue(self.callback_called)

    def test_current_download_speed(self):

        self.assertEqual(self.progress.calculate_current_download_speed(), 0)
        self.progress.add_single_progress(50)
        time.sleep(0.25)

        self.assertEqual(self.progress.calculate_current_download_speed(), 0)
        self.progress.add_single_progress(50)
        time.sleep(0.75)

        self.assertAlmostEqual(self.progress.calculate_current_download_speed(), 100, delta=5)

    def test_for_zero_division_errors(self):

        progress = Progress(0)
        self.assertEqual(progress.get_single_progress_percentage(), 0.0)
        self.assertEqual(progress.get_total_percentage(), 0.0)

    def test_finished_numbers(self):

        while self.progress.get_total_percentage() < 100.0:
            self.progress.set_single_progress_total(100)
            while self.progress.get_single_progress_percentage() < 100.0:
                self.progress.add_single_progress(1)
            self.progress.next_file()

        self.assertEqual(self.progress.get_total_percentage(), 100.0)
        self.assertEqual(self.progress.get_single_progress_percentage(), 100.0)
        self.assertEqual(self.progress.get_single_progress(), 100)
        self.assertEqual(self.progress.get_single_progress(), self.progress.get_single_progress_total())

    def test_going_past_total_limit(self):

        while self.progress.get_total_percentage() < 100.0:
            self.progress.next_file()

        self.assertEqual(self.progress.get_total_percentage(), 100.0)
        self.assertEqual(self.progress.total_progress, 10)
        self.progress.next_file()
        self.assertEqual(self.progress.get_total_percentage(), 100.0)
        self.assertEqual(self.progress.total_progress, 10)
        self.progress.next_file()
        self.assertEqual(self.progress.get_total_percentage(), 100.0)
        self.assertEqual(self.progress.total_progress, 10)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import unittest
import irc.events
from xdcc_dl.xdcc.layers.irc.IrcEventPrinter import IrcEventPrinter


class TestException(Exception):
    pass


class UnitTests(unittest.TestCase):

    def setUp(self):
        self.client = IrcEventPrinter

    def tearDown(self):
        self.client.quit()

    def test_defined_methods(self):

        class Tester(IrcEventPrinter):
            def on_welcome(self, conn, event):

                self.on_privnotice(conn, event)
                self.on_privmsg(conn, event)
                self.on_pubnotice(conn, event)
                self.on_pubmsg(conn, event)
                self.on_ping(conn, event)
                self.on_ctcp(conn, event)
                raise TestException()

        self.client = Tester("irc.namibsun.net", "random")
        try:
            self.client.start()
            self.assertTrue(False)
        except TestException:
            self.assertTrue(True)

    def test_auto_generated_methods(self):

        class Tester(IrcEventPrinter):

            assertions_true = True

            # noinspection PyUnusedLocal
            def on_welcome(self, conn, event):

                for event in irc.events.all:
                    self.assertions_true = self.assertions_true and callable(getattr(self, "on_" + event))
                raise TestException()

        self.client = Tester("irc.namibsun.net", "random")

        try:
            self.client.start()
            self.assertTrue(False)
        except TestException:
            self.assertTrue(self.client.assertions_true)

    def test_version_ctcp(self):

        class Tester(IrcEventPrinter):

            assertions_true = False

            def on_welcome(self, conn, event):

                event.arguments = ["VERSION"]
                self.on_ctcp(conn, event)

                raise TestException()

            def on_ctcp(self, conn, event):
                super().on_ctcp(conn, event)
                self.assertions_true = event.arguments[0] == "VERSION"

        self.client = Tester("irc.namibsun.net", "random")

        try:
            self.client.start()
            self.assertTrue(False)
        except TestException:
            self.assertTrue(self.client.assertions_true)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import time


class ConnectionStates(object):
    """
    Class that separates state variables from the other IRC/XDCC classes
    Layer -1 of the XDCC Bot

    All of these variables should be considered obsolete for each XDCC pack
    """

    def __init__(self):
        """
        Initializes the state variables
        """
        # Connection
        self.connected_to_server = False
        self.channel_joined = False
        self.channel_join_required = False

        # Download
        self.already_requested = False
        self.download_started = False
        self.dcc_resume_requested = False
        self.already_downloaded = False

        # Download File Specific
        self.current_pack = None
        self.file = None
        self.dcc_connection = None
        self.peer_address = None
        self.peer_port = None
        self.filesize = None

        # Stats
        self.start_time = time.time()

    def reset_connection_state(self) -> None:
        """
        Resets all connection state variables

        :return: None
        """
        # Connection
        self.connected_to_server = False
        self.channel_joined = False
        self.channel_join_required = False

        # Download
        self.already_requested = False
        self.download_started = False
        self.dcc_resume_requested = False
        self.already_downloaded = False

        # Download File Specific
        self.current_pack = None
        self.file = None
        self.dcc_connection = None
        self.peer_address = None
        self.peer_port = None
        self.filesize = None

        # Stats
        self.start_time = time.time()



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import unittest
from xdcc_dl.metadata import SentryLogger
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.IrcServer import IrcServer
from xdcc_dl.xdcc.layers.helpers.BotChannelMapper import BotChannelMapper
from xdcc_dl.xdcc.layers.xdcc.XDCCInitiator import XDCCInitiator, NoValidWhoisQueryException


class TestException(Exception):
    pass


class UnitTests(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_dcc_check(self):

        class Tester(XDCCInitiator):

            def on_welcome(self, conn, event):
                event.arguments = ["NOTDCC"]
                self.on_ctcp(conn, event)
                raise TestException()

        try:
            Tester("irc.namibsun.net", "random").start()
            self.assertTrue(False)
        except TestException:
            self.assertTrue(True)

    def test_missing_whois(self):
        class DummySentry(object):
            # noinspection PyPep8Naming
            def captureMessage(self, string):
                pass

        BotChannelMapper.bot_channel_map = {}
        SentryLogger.sentry = DummySentry()

        try:
            initiator = XDCCInitiator("irc.rizon.net", "random")
            initiator.current_pack = XDCCPack(IrcServer("irc.rizon.net"), "HelloKitty", 1)
            initiator.start()
            self.assertTrue(False)
        except NoValidWhoisQueryException:
            self.assertTrue(True)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""


class IrcServer(object):
    """
    Class that models an IRC server
    """

    def __init__(self, server_address: str, server_port: int = 6667) -> None:
        """
        Initializes the Server's information

        :param server_address: the address of the IRC Server
        :param server_port:    the port of the server, which defaults to 6667
                               and can usually safely stay that way
        """
        self.address = server_address
        self.port = server_port

    def get_address(self) -> str:
        """
        :return: the server address
        """
        return self.address

    def get_port(self) -> int:
        """
        :return: the server port
        """
        return self.port



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""
# imports
import sys
import unittest
from xdcc_dl.main import main


class UnitTests(unittest.TestCase):

    def setUp(self):
        sys.argv = [sys.argv[0]]

    def tearDown(self):
        sys.argv = [sys.argv[0]]

    def test_no_arguments(self):
        try:
            main()
            self.assertEqual(True, False)
        except SystemExit:
            self.assertEqual(True, True)

    def test_keyboard_interrupt(self):

        if sys.version_info[0] >= 3:
            exec("import builtins as __builtin__") in globals()

            # noinspection PyUnusedLocal
            def interrupt(arg):
                if arg == "No arguments passed. See --help for more details":
                    raise KeyboardInterrupt()
                else:
                    self.assertEqual("Thanks for using xdcc-downloader!", arg)

            exec("real_print = __builtin__.print") in globals()
            exec("__builtin__.print = interrupt") in globals()

            main()

            exec("__builtin__.print = real_print") in globals()



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import sys
import time
from threading import Thread
from xdcc_dl.entities.Progress import Progress
from xdcc_dl.pack_searchers.PackSearcher import PackSearcher
from xdcc_dl.entities.XDCCPack import xdcc_packs_from_xdcc_message
from xdcc_dl.gui.pyuic.xdcc_downloader import Ui_XDCCDownloaderWindow
from xdcc_dl.xdcc.MultipleServerDownloader import MultipleServerDownloader
from PyQt5.QtWidgets import QMainWindow, QApplication, QTreeWidgetItem, QPushButton, QMessageBox
from PyQt5.QtCore import pyqtSignal, Qt


class XDCCDownloaderGui(QMainWindow, Ui_XDCCDownloaderWindow):
    """
    Class that models a QT GUI for the XDCC Downloader
    """

    spinner_start_signal = pyqtSignal(str, name="spinner_start")
    spinner_updater_signal = pyqtSignal(QPushButton, str, name="spinner_updater")
    refresh_search_results_signal = pyqtSignal(str, name="refresh_search_results")
    refresh_download_queue_signal = pyqtSignal(str, name="refresh_download_queue")
    show_download_complete_message_signal = pyqtSignal(str, name="show_download_complete_message")
    progress_update_signal = pyqtSignal(float, float, name="progress_update")

    def __init__(self, parent: QMainWindow = None) -> None:
        """
        Sets up the interactive UI elements

        :param parent: the parent window
        """
        super().__init__(parent)
        self.setupUi(self)

        header = self.search_result_tree_widget.header()
        header.hideSection(header.logicalIndex(header.visualIndex(0)))

        self.spinner_start_signal.connect(lambda x: self.start_spinner(x))
        self.spinner_updater_signal.connect(lambda x, y: self.update_spinner(x, y))
        self.refresh_search_results_signal.connect(self.refresh_search_results)
        self.refresh_download_queue_signal.connect(self.refresh_download_queue)
        self.show_download_complete_message_signal.connect(self.show_download_complete_message)
        self.progress_update_signal.connect(self.progress_update)

        self.searching = False
        self.downloading = False
        self.downloader = None

        self.search_results = []
        self.download_queue = []

        for searcher in ["All"] + PackSearcher.get_available_pack_searchers():
            self.search_engine_combo_box.addItem(searcher)
        self.destination_edit.setText(os.getcwd())

        self.search_button.clicked.connect(self.search)
        self.download_button.clicked.connect(self.download)
        self.add_button.clicked.connect(self.add_pack)

        self.search_term_edit.returnPressed.connect(self.search)

        self.up_arrow_button.clicked.connect(lambda: self.move_packs(up=True))
        self.down_arrow_button.clicked.connect(lambda: self.move_packs(down=True))
        self.left_arrow_button.clicked.connect(self.remove_packs_from_queue)
        self.rigth_arrow_button.clicked.connect(self.add_packs_to_queue)

    def add_pack(self):

        message = self.message_edit.text()
        server = self.server_edit.text()

        self.download_queue += xdcc_packs_from_xdcc_message(message, server=server)
        self.refresh_download_queue()

    def search(self) -> None:
        """
        Starts the XDCC search

        :return: None
        """
        if not self.searching:

            self.searching = True

            search_term = self.search_term_edit.text()
            search_engine = self.search_engine_combo_box.currentText()
            search_engines = [search_engine] if search_engine != "All" else PackSearcher.get_available_pack_searchers()

            searcher = PackSearcher(search_engines)

            def search_thread():

                self.spinner_start_signal.emit("search")
                self.search_results = searcher.search(search_term)
                self.refresh_search_results_signal.emit("")
                self.searching = False

            Thread(target=search_thread).start()

    def refresh_search_results(self) -> None:
        """
        Refreshes the Search Results Tree Widget with the current search results

        :return: None
        """
        self.search_result_tree_widget.clear()
        self.search_results.sort(key=lambda x: x.get_bot())

        for i, result in enumerate(self.search_results):
            column = [str(i),
                      result.get_bot(), str(result.get_packnumber()), str(result.get_size()), result.get_filename()]
            self.search_result_tree_widget.addTopLevelItem(QTreeWidgetItem(column))

        self.search_result_tree_widget.sortByColumn(1, Qt.AscendingOrder)

    def start_spinner(self, spinner_type: str) -> None:
        """
        Starts a spinner animation while either searching or downloading

        :param spinner_type: The type of spinner (a string that's either 'download' or 'search')
        :return:         None
        """
        def spin():

            search = spinner_type == "search"
            download = spinner_type == "download"

            while self.searching or self.downloading:

                if self.searching and search:
                    new_text = "Searching" + (self.search_button.text().count(".") % 3 + 1) * "."
                    self.spinner_updater_signal.emit(self.search_button, new_text)

                if self.downloading and download:
                    new_text = "Downloading" + (self.download_button.text().count(".") % 3 + 1) * "."
                    self.spinner_updater_signal.emit(self.download_button, new_text)

                time.sleep(0.3)

            if search and not self.searching:
                self.search_button.setText("Search")
            if download and not self.downloading:
                self.download_button.setText("Download")

        Thread(target=spin).start()

    # noinspection PyMethodMayBeStatic
    def update_spinner(self, widget: QPushButton, text: str) -> None:
        """
        Sets the text of the given spinner button

        :param widget: The button to change the text of
        :param text:   The text to display
        :return:       None
        """
        widget.setText(text)

    def add_packs_to_queue(self):
        """
        Adds selected packs from the search results tree widget to the download queue

        :return: None
        """

        for item in self.search_result_tree_widget.selectedItems():
            self.download_queue.append(self.search_results[int(item.text(0))])

        self.refresh_download_queue()

    def remove_packs_from_queue(self) -> None:
        """
        Removes all selected elements from the Download Queue

        :return: None
        """
        rows_to_pop = []
        for row in self.download_queue_list_widget.selectedIndexes():
            rows_to_pop.append(row.row())

        for row in reversed(sorted(rows_to_pop)):
            self.download_queue.pop(row)

        self.refresh_download_queue()

    def move_packs(self, up: bool = False, down: bool = False) -> None:
        """
        Moves items on the queue up or down

        :param up:   Pushes the selected elements up
        :param down: Pushes the selected elements down
        :return:     None
        """

        size_check = (lambda x: x > 0) if up and not down else (lambda x: x < len(self.download_queue) - 1)
        index_change = (lambda x: x - 1) if up and not down else (lambda x: x + 1)

        indexes = self.download_queue_list_widget.selectedIndexes() if up and not down \
            else reversed(self.download_queue_list_widget.selectedIndexes())

        for row in indexes:

            index = row.row()
            if size_check(index):
                self.download_queue.insert(index_change(index), self.download_queue.pop(index))

        self.refresh_download_queue()

    def refresh_download_queue(self):
        """
        Reloads all elements currently in the download queue

        :return: None
        """
        self.download_queue_list_widget.clear()
        for pack in self.download_queue:
            self.download_queue_list_widget.addItem(pack.get_request_message(full=True))

    def download(self):
        """
        Starts the download of all packs in the download queue

        :return: None
        """
        if not self.downloading:

            directory = self.destination_edit.text()

            if not os.path.isdir(directory):

                msg = self.generate_message(QMessageBox.Warning, "Invalid Directory",
                                            "The entered directory is not valid", directory)

                if not sys.argv == [sys.argv[0], "-platform", "minimal"]:  # pragma: no cover
                    msg.exec_()

            else:

                self.downloading = True

                for pack in self.download_queue:
                    pack.set_directory(directory)

                def do_download():

                    progress = Progress(len(self.download_queue),
                                        callback=lambda a, b, sin, d, e, tot, g, h: self.progress_update_signal.emit(
                                            sin, tot))

                    self.spinner_start_signal.emit("download")
                    self.downloader = MultipleServerDownloader("random")
                    results = self.downloader.download(self.download_queue, progress)
                    self.download_queue = []
                    self.refresh_download_queue_signal.emit("")
                    self.progress_update_signal.emit(0.0, 0.0)
                    self.downloader.quit()
                    self.downloading = False

                    list_of_downloaded_packs = ""
                    for result in results:
                        list_of_downloaded_packs += result.get_filepath() + "\n"

                    self.show_download_complete_message_signal.emit(list_of_downloaded_packs)

                Thread(target=do_download).start()

    def progress_update(self, single_percentage: float, total_percentage: float):
        """
        Updates the progress bars

        :param single_percentage: The single completion percentage
        :param total_percentage:  The total completion percentage
        :return:                  None
        """
        self.single_progress_bar.setValue(single_percentage)
        self.total_progress_bar.setValue(total_percentage)

    # noinspection PyMethodMayBeStatic
    def generate_message(self, icon: int, title: str, text: str, detailed_text: str) -> QMessageBox:
        """
        Generates a Message Dialog

        :param icon:           The icon to display
        :param title:          The title to display
        :param text:           The primary text to display
        :param detailed_text:  The detailed text to display
        :return:               The generated message dialog
        """
        msg = QMessageBox()
        msg.setIcon(icon)
        msg.setWindowTitle(title)
        msg.setText(text)
        msg.setInformativeText(detailed_text)
        msg.setStandardButtons(QMessageBox.Ok)
        return msg

    def show_download_complete_message(self, details: str) -> None:
        """
        Message shown when the download has finished

        :param details: A formatted list of packs
        :return:        None
        """
        msg = self.generate_message(QMessageBox.Information, "Download Complete",
                                    "The download has been completed", "List of downloaded packs:")
        msg.setDetailedText(details.rstrip().lstrip())

        if not sys.argv == [sys.argv[0], "-platform", "minimal"]:  # pragma: no cover
            msg.exec_()


def start():  # pragma: no cover
    """
    Starts the Start Page GUI

    :return: None
    """
    app = QApplication(sys.argv)
    form = XDCCDownloaderGui()
    form.show()
    app.exec_()
    form.searching = False
    form.downloading = False
    if form.downloader is not None:
        form.downloader.quit()



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
from typing import List, Dict

from xdcc_dl.logging import Logger
from xdcc_dl.entities.User import User
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.Progress import Progress
from xdcc_dl.entities.IrcServer import IrcServer
from xdcc_dl.xdcc.XDCCDownloader import XDCCDownloader


class MultipleServerDownloader(object):
    """
    Class that can handle downloading from various server sources
    """

    def __init__(self, user: User or str, logger: Logger or int = 0):
        """
        Creates a new multiple server downloader, without specifying the server.

        :param user:    The user with which to log in
        :param logger:  The logger to use
        """
        self.user = user
        self.logger = logger
        self.current_downloader = None
        self.quit_called = False

    def download(self, packs: List[XDCCPack], progress: Progress = None) -> Dict[XDCCPack, str]:
        """
        Downloads all XDCC packs specified. Optionally shares state with other threads using a Progress object

        :param packs:    The packs to download
        :param progress: Optional Progress object
        :return:         Dictionary of packs mapped to status codes:
                         "OK":           Download was successful
                         "BOTNOTFOUND":  Bot was not found
                         "NETWORKERROR": Download failed due to network error
                         "INCOMPLETE":   Download was incomplete
                         "EXISTED":      File already existed and was completely downloaded
        """
        results = {}

        packservers = {}

        for pack in packs:
            try:
                packservers[pack.get_server().get_address()].append(pack)
            except KeyError:
                packservers[pack.get_server().get_address()] = [pack]

        for server in packservers:

            if not self.quit_called:
                self.current_downloader = XDCCDownloader(IrcServer(server), self.user, self.logger)
                server_results = self.current_downloader.download(packservers[server], progress)

                for result in server_results:
                    results[result] = server_results[result]

        return results

    def quit(self) -> None:
        """
        Quits the current downloader and stops all downloads to come

        :return: None
        """
        if self.current_downloader is not None:
            self.current_downloader.quit()
        self.quit_called = True



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import shlex
import irc.client
from typing import List
from xdcc_dl.metadata import SentryLogger
# noinspection PyPep8Naming
from xdcc_dl.logging.LoggingTypes import LoggingTypes as LOG
from xdcc_dl.xdcc.layers.xdcc.MessageSender import MessageSender


class IncorrectFileSentException(Exception):
    """
    Gets raised whenever the bot sends the incorrect, or at least not-predicted file
    """
    pass


class NoValidWhoisQueryException(Exception):
    """
    Gets raised when a bot is encountered that requires the user to join a specific channel, but does not
    provide the identity of that channel via whois
    """


# noinspection PyUnusedLocal
class XDCCInitiator(MessageSender):
    """
    Initiates the XDCC Connection.
    Layer 4 of the XDCC Bot
    """

    def on_ctcp(self, connection: irc.client.ServerConnection, event: irc.client.Event):
        """
        Client-to-Client Connection which initiates the XDCC handshake

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        super().on_ctcp(connection, event)

        if event.arguments[0] != "DCC":
            return

        payload = shlex.split(event.arguments[1])

        if payload[0] == "SEND":
            self.dcc_send_handler(payload, connection)
        elif payload[0] == "ACCEPT":
            self.dcc_accept_handler(payload, connection)
        else:  # pragma: no cover
            return

    def dcc_send_handler(self, ctcp_arguments: List[str], connection: irc.client.ServerConnection) -> None:
        """
        Handles incoming CTCP DCC SENDs. Initiates a download or RESUME request.

        :param ctcp_arguments: The CTCP Arguments
        :param connection:     The connection to use for DCC connections
        :raises:               AlreadyDownloaded, if the file was already downloaded
        :return:               None
        """
        self.logger.log("Handling DCC SEND Handshake", LOG.DCC_SEND_HANDSHAKE)

        filename = ctcp_arguments[1]
        if not self.current_pack.is_filename_valid(filename):
            self.logger.log("Incorrect file sent", LOG.INCORRECT_FILE)
            raise IncorrectFileSentException()

        self.peer_address = irc.client.ip_numstr_to_quad(ctcp_arguments[2])
        self.peer_port = int(ctcp_arguments[3])
        self.filesize = int(ctcp_arguments[4])

        self.progress.set_single_progress_total(int(self.filesize))
        self.current_pack.set_filename(filename)

        if os.path.exists(self.current_pack.get_filepath()) and not self.dcc_resume_requested:

            position = os.path.getsize(self.current_pack.get_filepath())

            if position >= self.filesize:

                self.logger.log("File already completely downloaded. Aborting", LOG.DOWNLOAD_WAS_DONE)
                self.already_downloaded = True
                self.dcc_connection = self.dcc_connect(self.peer_address, self.peer_port, "raw")

            else:

                self.logger.log("Requesting DCC RESUME", LOG.DCC_RESUME_REQUEST)
                self.progress.set_single_progress(position)

                self.dcc_resume_requested = True  # Let bot know that resume was attempted
                resume_parameter = "\"" + filename + "\" " + str(self.peer_port) + " " + str(position)

                # -> on_ctcp -> dcc_accept_handler (Or dcc_send_handler if resume fails)
                connection.ctcp("DCC RESUME", self.current_pack.get_bot(), resume_parameter)

        else:

            if self.dcc_resume_requested:
                self.logger.log("DCC Resume Failed. Starting from scratch.", LOG.DCC_RESUME_FAILED)
                os.remove(self.current_pack.get_filepath())
                self.progress.set_single_progress(0)

            self.logger.log("Starting Download of " + filename, LOG.DOWNLOAD_START)

            self.file = open(self.current_pack.get_filepath(), "wb")
            self.dcc_connection = self.dcc_connect(self.peer_address, self.peer_port, "raw")  # -> on_dccmsg
            self.download_started = True

    def dcc_accept_handler(self, ctcp_arguments: List[str], connection: irc.client.ServerConnection) -> None:
        """
        Handles DCC ACCEPT messages. Resumes a download.

        :param ctcp_arguments: The CTCP arguments
        :param connection:     The connection to use for DCC connections
        :return:               None
        """
        self.logger.log("DCC RESUME request successful", LOG.DCC_RESUME_SUCCESS)
        self.logger.log("Resuming Download of " + self.current_pack.get_filepath(), LOG.DOWNLOAD_RESUME)

        self.file = open(self.current_pack.get_filepath(), "ab")
        self.dcc_connection = self.dcc_connect(self.peer_address, self.peer_port, "raw")  # -> on_dccmsg
        self.download_started = True

    def on_privnotice(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Checks if a private notice is sent by the sending bot that a pack was already requested.
        If that is the case, another attempt will be made on the next ping.
        Also checks for channel-based XDCC denies

        :param connection: the IRC connection
        :param event:      the IRC event
        :raises:           NoValidWhoisQuery, if the bot told us that no correct channel was joined
        :return:           None
        """
        try:
            if "You already requested that pack" in event.arguments[0]:
                self.logger.log("Pack already requested, waiting for next Ping", LOG.ALREADY_REQUESTED)
                self.already_requested = True
            elif "You will have to re-send that, to the bot that transferred the file." in event.arguments[0]:
                connection.privmsg(self.current_pack.get_bot(), self.current_pack.get_request_message())
            elif "** XDCC SEND denied, you must be on a known channel to request a pack" in event.arguments[0]:
                SentryLogger.sentry.captureMessage("Did not join correct channel: Bot=" + self.current_pack.get_bot())
                raise NoValidWhoisQueryException()
            else:
                super().on_privnotice(connection, event)
        except IndexError:
            pass

    def on_ping(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Retries a download on a ping signal
        :param connection:
        :param event:
        :return:
        """
        super().on_ping(connection, event)
        if self.already_requested:
            log_message = "Resending XDCC Request to " + self.current_pack.get_bot()
            log_message += " for pack " + str(self.current_pack.get_packnumber())
            self.logger.log(log_message, LOG.MESSAGE_RETRY)

            self.already_requested = False
            connection.privmsg(self.current_pack.get_bot(), self.current_pack.get_request_message())  # -> on_ctcp
